//! Everything in this file is part of the public JSON api
//!
//! Adding methods on types is fine, but modifying structure breaks compatibility

use uuid::Uuid;

#[derive(
    Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash, serde::Deserialize, serde::Serialize,
)]
#[serde(transparent)]
pub(crate) struct GameId {
    inner: Uuid,
}

#[derive(
    Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash, serde::Deserialize, serde::Serialize,
)]
pub(crate) struct Piece {
    pub(crate) kind: PieceKind,
    pub(crate) color: Color,
}

#[derive(
    Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash, serde::Deserialize, serde::Serialize,
)]
pub(crate) struct Coordinates {
    pub(crate) file: File,
    pub(crate) rank: Rank,
}

#[derive(serde::Deserialize, serde::Serialize)]
#[serde(rename_all = "camelCase")]
pub(crate) struct Move {
    pub(crate) piece: Piece,
    pub(crate) from: Coordinates,
    pub(crate) to: Coordinates,
    pub(crate) game_id: GameId,
}

#[derive(serde::Deserialize, serde::Serialize)]
#[serde(rename_all = "camelCase")]
pub(crate) struct Start {
    pub(crate) player_color: Color,
}

#[derive(serde::Serialize)]
#[serde(rename_all = "camelCase")]
pub(crate) struct GameStart {
    pub(crate) board: Vec<(File, Rank, PieceKind, Color)>,
    pub(crate) game_id: GameId,
}

#[derive(
    Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash, serde::Deserialize, serde::Serialize,
)]
#[serde(rename_all = "camelCase")]
pub(crate) enum Color {
    Black,
    White,
}

#[derive(
    Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash, serde::Deserialize, serde::Serialize,
)]
#[serde(rename_all = "camelCase")]
pub(crate) enum PieceKind {
    Pawn,
    Rook,
    Knight,
    Bishop,
    Queen,
    King,
}

#[derive(
    Clone,
    Copy,
    Debug,
    PartialEq,
    Eq,
    PartialOrd,
    Ord,
    Hash,
    serde_repr::Deserialize_repr,
    serde_repr::Serialize_repr,
)]
#[repr(u8)]
pub(crate) enum Rank {
    One = 1,
    Two = 2,
    Three = 3,
    Four = 4,
    Five = 5,
    Six = 6,
    Seven = 7,
    Eight = 8,
}

#[derive(
    Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash, serde::Deserialize, serde::Serialize,
)]
#[serde(rename_all = "camelCase")]
pub(crate) enum File {
    A,
    B,
    C,
    D,
    E,
    F,
    G,
    H,
}

impl GameId {
    pub(crate) fn new() -> Self {
        GameId {
            inner: Uuid::new_v4(),
        }
    }
}

impl Color {
    pub(crate) fn is_white(self) -> bool {
        matches!(self, Self::White)
    }

    pub(crate) fn is_black(self) -> bool {
        matches!(self, Self::Black)
    }
}

impl Rank {
    pub(crate) fn next(self) -> Option<Self> {
        match self {
            Self::One => Some(Self::Two),
            Self::Two => Some(Self::Three),
            Self::Three => Some(Self::Four),
            Self::Four => Some(Self::Five),
            Self::Five => Some(Self::Six),
            Self::Six => Some(Self::Seven),
            Self::Seven => Some(Self::Eight),
            Self::Eight => None,
        }
    }

    pub(crate) fn prev(self) -> Option<Self> {
        match self {
            Self::One => None,
            Self::Two => Some(Self::One),
            Self::Three => Some(Self::Two),
            Self::Four => Some(Self::Three),
            Self::Five => Some(Self::Four),
            Self::Six => Some(Self::Five),
            Self::Seven => Some(Self::Six),
            Self::Eight => Some(Self::Seven),
        }
    }

    pub(crate) fn diff(self, other: Self) -> Option<u8> {
        if self > other {
            Some(self.to_number() - other.to_number())
        } else {
            None
        }
    }

    pub(crate) fn absolute_diff(self, other: Self) -> u8 {
        self.diff(other).or_else(|| other.diff(self)).unwrap_or(0)
    }

    pub(crate) fn to_number(self) -> u8 {
        match self {
            Self::One => 0,
            Self::Two => 1,
            Self::Three => 2,
            Self::Four => 3,
            Self::Five => 4,
            Self::Six => 5,
            Self::Seven => 6,
            Self::Eight => 7,
        }
    }

    pub(crate) fn from_number(value: u8) -> Option<Self> {
        match value {
            0 => Some(Self::One),
            1 => Some(Self::Two),
            2 => Some(Self::Three),
            3 => Some(Self::Four),
            4 => Some(Self::Five),
            5 => Some(Self::Six),
            6 => Some(Self::Seven),
            7 => Some(Self::Eight),
            _ => None,
        }
    }
}

impl File {
    pub(crate) fn next(self) -> Option<Self> {
        match self {
            Self::A => Some(Self::B),
            Self::B => Some(Self::C),
            Self::C => Some(Self::D),
            Self::D => Some(Self::E),
            Self::E => Some(Self::F),
            Self::F => Some(Self::G),
            Self::G => Some(Self::H),
            Self::H => None,
        }
    }

    pub(crate) fn prev(self) -> Option<Self> {
        match self {
            Self::A => None,
            Self::B => Some(Self::A),
            Self::C => Some(Self::B),
            Self::D => Some(Self::C),
            Self::E => Some(Self::D),
            Self::F => Some(Self::E),
            Self::G => Some(Self::F),
            Self::H => Some(Self::G),
        }
    }

    pub(crate) fn diff(self, other: Self) -> Option<u8> {
        if self > other {
            Some(self.to_number() - other.to_number())
        } else {
            None
        }
    }

    pub(crate) fn absolute_diff(self, other: Self) -> u8 {
        self.diff(other).or_else(|| other.diff(self)).unwrap_or(0)
    }

    pub(crate) fn to_number(self) -> u8 {
        match self {
            Self::A => 0,
            Self::B => 1,
            Self::C => 2,
            Self::D => 3,
            Self::E => 4,
            Self::F => 5,
            Self::G => 6,
            Self::H => 7,
        }
    }

    pub(crate) fn from_number(value: u8) -> Option<Self> {
        match value {
            0 => Some(Self::A),
            1 => Some(Self::B),
            2 => Some(Self::C),
            3 => Some(Self::D),
            4 => Some(Self::E),
            5 => Some(Self::F),
            6 => Some(Self::G),
            7 => Some(Self::H),
            _ => None,
        }
    }
}

impl std::fmt::Display for GameId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        std::fmt::Display::fmt(&self.inner, f)
    }
}
