use crate::api_types::{Color, Coordinates, File, Move, Piece, PieceKind, Rank};
use std::collections::HashMap;

#[derive(Clone, Debug)]
struct CastleState {
    can_king_castle: bool,
    can_queen_castle: bool,
}

#[derive(Default)]
struct SideEffects {
    en_passant: Option<Coordinates>,
    king_castle_invalidated: bool,
    queen_castle_invalidated: bool,
}

#[derive(Clone, Debug, Default)]
pub(crate) struct BoardState {
    inner: HashMap<Coordinates, Piece>,
    white_state: CastleState,
    black_state: CastleState,
    en_passant: Option<Coordinates>,
}

impl BoardState {
    pub(crate) fn starting_positions() -> Self {
        let mut this = Self::default();

        for (file, rank, kind, color) in INITIAL_PIECES {
            this.inner
                .insert(Coordinates { file, rank }, Piece { kind, color });
        }

        this
    }

    pub(crate) fn to_serializable(&self) -> Vec<(File, Rank, PieceKind, Color)> {
        self.inner
            .iter()
            .map(|(coordinates, piece)| {
                (coordinates.file, coordinates.rank, piece.kind, piece.color)
            })
            .collect()
    }

    pub(crate) fn handle(&mut self, action: Move) {
        let mut proposed = self.clone();

        let moving_color = action.piece.color;

        if let Some(piece) = proposed.inner.remove(&action.from) {
            // not allowed to move mismatched piece
            if piece != action.piece {
                return;
            }

            let opt = piece.can_move_to(&action.from, &action.to, &proposed);

            if let Some(side_effects) = opt {
                proposed.apply_side_effects(moving_color, side_effects);
                proposed.inner.insert(action.to, piece);
            } else {
                proposed.inner.insert(action.from, piece);
            }
        }

        if proposed.is_in_check(moving_color) {
            return;
        }

        *self = proposed;
    }

    fn apply_side_effects(&mut self, color: Color, side_effects: SideEffects) {
        self.en_passant = side_effects.en_passant;

        if color.is_white() {
            if side_effects.king_castle_invalidated {
                self.white_state.can_king_castle = false;
            }
            if side_effects.queen_castle_invalidated {
                self.white_state.can_queen_castle = false;
            }
        } else {
            if side_effects.king_castle_invalidated {
                self.black_state.can_king_castle = false;
            }
            if side_effects.queen_castle_invalidated {
                self.black_state.can_queen_castle = false;
            }
        }
    }

    fn is_in_check(&self, color: Color) -> bool {
        let king = Piece {
            kind: PieceKind::King,
            color,
        };

        let king_opt = self.inner.iter().find(|(_, piece)| **piece == king);

        if let Some((king_coords, _)) = king_opt {
            self.inner
                .iter()
                .filter(|(_, piece)| piece.color != color)
                .any(|(coords, piece)| piece.can_move_to(coords, king_coords, self).is_some())
        } else {
            false
        }
    }
}

impl Piece {
    fn can_move_to(
        &self,
        from: &Coordinates,
        to: &Coordinates,
        state: &BoardState,
    ) -> Option<SideEffects> {
        match self.kind {
            PieceKind::Pawn if self.color.is_white() && to.rank > from.rank => {
                self.check_pawn(from, to, state, Rank::Two, Rank::next)
            }
            PieceKind::Pawn if self.color.is_black() && to.rank < from.rank => {
                self.check_pawn(from, to, state, Rank::Seven, Rank::prev)
            }
            _ => None,
        }
    }

    fn check_pawn(
        &self,
        from: &Coordinates,
        to: &Coordinates,
        state: &BoardState,
        en_passant_rank: Rank,
        forward: impl Fn(Rank) -> Option<Rank>,
    ) -> Option<SideEffects> {
        let rank_diff = to.rank.absolute_diff(from.rank);
        let file_diff = to.file.absolute_diff(from.file);

        let mut side_effects = SideEffects::default();

        // move forward 1 or two steps as long as not blocked
        if file_diff == 0 && ((rank_diff == 2 && from.rank == en_passant_rank) || rank_diff == 1) {
            let mut path_rank = from.rank;

            while let Some(rank) = forward(path_rank) {
                let block_check = Coordinates {
                    rank,
                    file: to.file,
                };

                if state.inner.contains_key(&block_check) {
                    return None;
                }

                if rank == to.rank {
                    break;
                }

                // set en_passant to cell directly behind pawn
                side_effects.en_passant = Some(block_check);

                path_rank = rank;
            }

            return Some(side_effects);
        }

        // pawn capture move
        if file_diff == 1 && rank_diff == 1 {
            // check for real capture
            if let Some(piece) = state.inner.get(to) {
                if piece.color != self.color {
                    return Some(side_effects);
                } else {
                    return None;
                }
            }

            // check en_passant
            if let Some(coords) = &state.en_passant {
                if coords == to {
                    return Some(side_effects);
                }
            }
        }

        None
    }
}

impl Default for CastleState {
    fn default() -> Self {
        Self {
            can_king_castle: true,
            can_queen_castle: true,
        }
    }
}

const INITIAL_PIECES: [(File, Rank, PieceKind, Color); 32] = [
    (File::A, Rank::One, PieceKind::Rook, Color::White),
    (File::B, Rank::One, PieceKind::Knight, Color::White),
    (File::C, Rank::One, PieceKind::Bishop, Color::White),
    (File::D, Rank::One, PieceKind::Queen, Color::White),
    (File::E, Rank::One, PieceKind::King, Color::White),
    (File::F, Rank::One, PieceKind::Bishop, Color::White),
    (File::G, Rank::One, PieceKind::Knight, Color::White),
    (File::H, Rank::One, PieceKind::Rook, Color::White),
    (File::A, Rank::Two, PieceKind::Pawn, Color::White),
    (File::B, Rank::Two, PieceKind::Pawn, Color::White),
    (File::C, Rank::Two, PieceKind::Pawn, Color::White),
    (File::D, Rank::Two, PieceKind::Pawn, Color::White),
    (File::E, Rank::Two, PieceKind::Pawn, Color::White),
    (File::F, Rank::Two, PieceKind::Pawn, Color::White),
    (File::G, Rank::Two, PieceKind::Pawn, Color::White),
    (File::H, Rank::Two, PieceKind::Pawn, Color::White),
    (File::A, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::B, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::C, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::D, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::E, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::F, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::G, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::H, Rank::Seven, PieceKind::Pawn, Color::Black),
    (File::A, Rank::Eight, PieceKind::Rook, Color::Black),
    (File::B, Rank::Eight, PieceKind::Knight, Color::Black),
    (File::C, Rank::Eight, PieceKind::Bishop, Color::Black),
    (File::D, Rank::Eight, PieceKind::Queen, Color::Black),
    (File::E, Rank::Eight, PieceKind::King, Color::Black),
    (File::F, Rank::Eight, PieceKind::Bishop, Color::Black),
    (File::G, Rank::Eight, PieceKind::Knight, Color::Black),
    (File::H, Rank::Eight, PieceKind::Rook, Color::Black),
];
