use actix_cors::Cors;
use actix_web::{
    web::{Data, Json},
    App, HttpResponse, HttpServer,
};
use std::{
    collections::HashMap,
    sync::{Arc, Mutex},
    time::Duration,
};
use tokio::sync::mpsc::Sender;
use tracing::Instrument;
use tracing_actix_web::TracingLogger;

mod api_types;
mod board_state;
mod init_tracing;

use api_types::{GameId, GameStart, Move, Start};
use board_state::BoardState;

struct Reply {
    board_state: BoardState,
}

struct MoveSession {
    action: Move,
    reply: tokio::sync::oneshot::Sender<Reply>,
}

#[derive(Clone, Debug)]
struct GameData {
    sender: Sender<MoveSession>,
}

#[derive(Clone, Default)]
struct GameState {
    inner: Arc<Mutex<HashMap<GameId, GameData>>>,
}

struct GameDropper(GameState, GameId);

impl GameState {
    fn data_for_id(&self, game_id: &GameId) -> Option<GameData> {
        self.inner.lock().unwrap().get(game_id).cloned()
    }

    fn save_data(&self, game_id: GameId, data: GameData) {
        self.inner.lock().unwrap().insert(game_id, data);
    }
}

impl Drop for GameDropper {
    fn drop(&mut self) {
        self.0.inner.lock().unwrap().remove(&self.1);
    }
}

async fn start(Json(start): Json<Start>, game_state: Data<GameState>) -> HttpResponse {
    let game_id = GameId::new();

    let game_dropper = GameDropper((**game_state).clone(), game_id.clone());

    let (sender, mut rx) = tokio::sync::mpsc::channel(1);

    game_state.save_data(game_id.clone(), GameData { sender });

    let game_span = tracing::info_span!(
        parent: None,
        "Game Session",
        game.id = %game_id,
    );

    let mut board_state = BoardState::starting_positions();
    let serializable = board_state.to_serializable();

    actix_web::rt::spawn(
        async move {
            while let Ok(Some(msg)) =
                actix_web::rt::time::timeout(Duration::from_secs(60 * 5), rx.recv()).await
            {
                board_state.handle(msg.action);

                let _ = msg.reply.send(Reply {
                    board_state: board_state.clone(),
                });
            }

            drop(game_dropper);
        }
        .instrument(game_span),
    );

    HttpResponse::Ok().json(GameStart {
        board: serializable,
        game_id,
    })
}

async fn make_move(Json(action): Json<Move>, game_state: Data<GameState>) -> HttpResponse {
    if let Some(data) = game_state.data_for_id(&action.game_id) {
        let (reply, rx) = tokio::sync::oneshot::channel();

        let res = data.sender.send(MoveSession { action, reply }).await;

        if res.is_err() {
            return HttpResponse::InternalServerError().finish();
        }

        if let Ok(reply) = rx.await {
            return HttpResponse::Ok().json(reply.board_state.to_serializable());
        }

        return HttpResponse::InternalServerError().finish();
    }

    HttpResponse::BadRequest().finish()
}

#[actix_web::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    init_tracing::init_tracing()?;

    let game_state = GameState::default();

    HttpServer::new(move || {
        App::new()
            .app_data(Data::new(game_state.clone()))
            .wrap(TracingLogger::default())
            .wrap(
                Cors::default()
                    .allow_any_origin()
                    .allow_any_header()
                    .allow_any_method(),
            )
            .route("/start", actix_web::web::post().to(start))
            .route("/move", actix_web::web::post().to(make_move))
    })
    .bind("0.0.0.0:8000")?
    .run()
    .await?;

    Ok(())
}
